/*
  This file is part of TALER
  Copyright (C) 2017-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_post-recoup-refresh.c
 * @brief Implementation of the /recoup-refresh request of the exchange's HTTP API
 * @author Christian Grothoff
 */
#include "taler/platform.h"
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_json_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler/taler_json_lib.h"
#include "taler/taler_exchange_service.h"
#include "exchange_api_common.h"
#include "exchange_api_handle.h"
#include "taler/taler_signatures.h"
#include "exchange_api_curl_defaults.h"


/**
 * @brief A Recoup-Refresh Handle
 */
struct TALER_EXCHANGE_PostRecoupRefreshHandle
{

  /**
   * The base URL for this request.
   */
  char *base_url;

  /**
   * The full URL for this request, set during _start.
   */
  char *url;

  /**
   * Minor context that holds body and headers.
   */
  struct TALER_CURL_PostContext post_ctx;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_PostRecoupRefreshCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_EXCHANGE_POST_RECOUP_REFRESH_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * The keys of the exchange this request handle will use.
   */
  struct TALER_EXCHANGE_Keys *keys;

  /**
   * Public key of the coin we are trying to get paid back.
   */
  struct TALER_CoinSpendPublicKeyP coin_pub;

  /**
   * Pre-built request body.
   */
  json_t *body;

};


/**
 * Parse a recoup-refresh response.  If it is valid, call the callback.
 *
 * @param ph recoup-refresh handle
 * @param json json reply with the signature
 * @return #GNUNET_OK if the signature is valid and we called the callback;
 *         #GNUNET_SYSERR if not (callback must still be called)
 */
static enum GNUNET_GenericReturnValue
process_recoup_response (
  const struct TALER_EXCHANGE_PostRecoupRefreshHandle *ph,
  const json_t *json)
{
  struct TALER_EXCHANGE_PostRecoupRefreshResponse rrr = {
    .hr.reply = json,
    .hr.http_status = MHD_HTTP_OK
  };
  struct GNUNET_JSON_Specification spec_refresh[] = {
    GNUNET_JSON_spec_fixed_auto ("old_coin_pub",
                                 &rrr.details.ok.old_coin_pub),
    GNUNET_JSON_spec_end ()
  };

  if (GNUNET_OK !=
      GNUNET_JSON_parse (json,
                         spec_refresh,
                         NULL, NULL))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  ph->cb (ph->cb_cls,
          &rrr);
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP /recoup-refresh request.
 *
 * @param cls the `struct TALER_EXCHANGE_PostRecoupRefreshHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_recoup_refresh_finished (void *cls,
                                long response_code,
                                const void *response)
{
  struct TALER_EXCHANGE_PostRecoupRefreshHandle *ph = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_PostRecoupRefreshResponse rrr = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  ph->job = NULL;
  switch (response_code)
  {
  case 0:
    rrr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    if (GNUNET_OK !=
        process_recoup_response (ph,
                                 j))
    {
      GNUNET_break_op (0);
      rrr.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
      rrr.hr.http_status = 0;
      break;
    }
    TALER_EXCHANGE_post_recoup_refresh_cancel (ph);
    return;
  case MHD_HTTP_BAD_REQUEST:
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_FORBIDDEN:
    /* Nothing really to verify, exchange says one of the signatures is
       invalid; as we checked them, this should never happen, we
       should pass the JSON reply to the application */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_NOT_FOUND:
    /* Nothing really to verify, this should never
       happen, we should pass the JSON reply to the application */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_CONFLICT:
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_GONE:
    /* Kind of normal: the money was already sent to the merchant
       (it was too late for the refund). */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  default:
    /* unexpected response code */
    rrr.hr.ec = TALER_JSON_get_error_code (j);
    rrr.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for exchange recoup-refresh\n",
                (unsigned int) response_code,
                (int) rrr.hr.ec);
    GNUNET_break (0);
    break;
  }
  ph->cb (ph->cb_cls,
          &rrr);
  TALER_EXCHANGE_post_recoup_refresh_cancel (ph);
}


struct TALER_EXCHANGE_PostRecoupRefreshHandle *
TALER_EXCHANGE_post_recoup_refresh_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  struct TALER_EXCHANGE_Keys *keys,
  const struct TALER_EXCHANGE_DenomPublicKey *pk,
  const struct TALER_DenominationSignature *denom_sig,
  const struct TALER_ExchangeBlindingValues *exchange_vals,
  const struct TALER_PublicRefreshMasterSeedP *rms,
  const struct TALER_PlanchetMasterSecretP *ps,
  unsigned int idx)
{
  struct TALER_EXCHANGE_PostRecoupRefreshHandle *ph;
  struct TALER_DenominationHashP h_denom_pub;
  struct TALER_CoinSpendPrivateKeyP coin_priv;
  union GNUNET_CRYPTO_BlindingSecretP bks;
  struct TALER_CoinSpendSignatureP coin_sig;

  (void) rms; // FIXME: why did we pass this again?
  (void) idx; // FIXME: why did we pass this again?
  ph = GNUNET_new (struct TALER_EXCHANGE_PostRecoupRefreshHandle);
  ph->ctx = ctx;
  ph->base_url = GNUNET_strdup (url);
  ph->keys = TALER_EXCHANGE_keys_incref (keys);
  TALER_planchet_setup_coin_priv (ps,
                                  exchange_vals,
                                  &coin_priv);
  TALER_planchet_blinding_secret_create (ps,
                                         exchange_vals,
                                         &bks);
  GNUNET_CRYPTO_eddsa_key_get_public (&coin_priv.eddsa_priv,
                                      &ph->coin_pub.eddsa_pub);
  TALER_denom_pub_hash (&pk->key,
                        &h_denom_pub);
  TALER_wallet_recoup_refresh_sign (&h_denom_pub,
                                    &bks,
                                    &coin_priv,
                                    &coin_sig);
  ph->body = GNUNET_JSON_PACK (
    GNUNET_JSON_pack_data_auto ("denom_pub_hash",
                                &h_denom_pub),
    TALER_JSON_pack_denom_sig ("denom_sig",
                               denom_sig),
    TALER_JSON_pack_exchange_blinding_values ("ewv",
                                              exchange_vals),
    GNUNET_JSON_pack_data_auto ("coin_sig",
                                &coin_sig),
    GNUNET_JSON_pack_data_auto ("coin_blind_key_secret",
                                &bks));
  switch (denom_sig->unblinded_sig->cipher)
  {
  case GNUNET_CRYPTO_BSA_INVALID:
    json_decref (ph->body);
    GNUNET_free (ph->base_url);
    TALER_EXCHANGE_keys_decref (ph->keys);
    GNUNET_free (ph);
    GNUNET_break (0);
    return NULL;
  case GNUNET_CRYPTO_BSA_RSA:
    break;
  case GNUNET_CRYPTO_BSA_CS:
    break;
  }
  return ph;
}


enum TALER_ErrorCode
TALER_EXCHANGE_post_recoup_refresh_start (
  struct TALER_EXCHANGE_PostRecoupRefreshHandle *ph,
  TALER_EXCHANGE_PostRecoupRefreshCallback cb,
  TALER_EXCHANGE_POST_RECOUP_REFRESH_RESULT_CLOSURE *cb_cls)
{
  CURL *eh;
  char arg_str[sizeof (struct TALER_CoinSpendPublicKeyP) * 2 + 32];

  ph->cb = cb;
  ph->cb_cls = cb_cls;
  {
    char pub_str[sizeof (struct TALER_CoinSpendPublicKeyP) * 2];
    char *end;

    end = GNUNET_STRINGS_data_to_string (
      &ph->coin_pub,
      sizeof (struct TALER_CoinSpendPublicKeyP),
      pub_str,
      sizeof (pub_str));
    *end = '\0';
    GNUNET_snprintf (arg_str,
                     sizeof (arg_str),
                     "coins/%s/recoup-refresh",
                     pub_str);
  }
  ph->url = TALER_url_join (ph->base_url,
                            arg_str,
                            NULL);
  if (NULL == ph->url)
  {
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Could not construct request URL.\n");
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  }
  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
              "URL for recoup-refresh: `%s'\n",
              ph->url);
  eh = TALER_EXCHANGE_curl_easy_get_ (ph->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&ph->post_ctx,
                              eh,
                              ph->body)) )
  {
    GNUNET_break (0);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  ph->job = GNUNET_CURL_job_add2 (ph->ctx,
                                  eh,
                                  ph->post_ctx.headers,
                                  &handle_recoup_refresh_finished,
                                  ph);
  if (NULL == ph->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_post_recoup_refresh_cancel (
  struct TALER_EXCHANGE_PostRecoupRefreshHandle *ph)
{
  if (NULL != ph->job)
  {
    GNUNET_CURL_job_cancel (ph->job);
    ph->job = NULL;
  }
  TALER_curl_easy_post_finished (&ph->post_ctx);
  GNUNET_free (ph->url);
  GNUNET_free (ph->base_url);
  json_decref (ph->body);
  TALER_EXCHANGE_keys_decref (ph->keys);
  GNUNET_free (ph);
}


/* end of exchange_api_post-recoup-refresh.c */
